"use strict";
/* eslint-disable @typescript-eslint/no-explicit-any */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DefaultPrismaProxyHandler = void 0;
exports.makeProxy = makeProxy;
const constants_1 = require("../../constants");
const cross_1 = require("../../cross");
const promise_1 = require("./promise");
/**
 * A default implementation of @see PrismaProxyHandler which directly
 * delegates to the wrapped Prisma client. It offers a few overridable
 * methods to allow more easily inject custom logic.
 */
class DefaultPrismaProxyHandler {
    constructor(prisma, model, options) {
        this.prisma = prisma;
        this.model = model;
        this.options = options;
    }
    withFluentCall(method, args, postProcess = true) {
        args = args ? (0, cross_1.clone)(args) : {};
        const promise = (0, promise_1.createFluentPromise)(() => __awaiter(this, void 0, void 0, function* () {
            args = yield this.preprocessArgs(method, args);
            const r = yield this.prisma[this.model][method](args);
            return postProcess ? this.processResultEntity(method, r) : r;
        }), args, this.options.modelMeta, this.model);
        return promise;
    }
    deferred(method, args, postProcess = true) {
        return (0, promise_1.createDeferredPromise)(() => __awaiter(this, void 0, void 0, function* () {
            args = yield this.preprocessArgs(method, args);
            const r = yield this.prisma[this.model][method](args);
            return postProcess ? this.processResultEntity(method, r) : r;
        }));
    }
    findUnique(args) {
        return this.withFluentCall('findUnique', args);
    }
    findUniqueOrThrow(args) {
        return this.withFluentCall('findUniqueOrThrow', args);
    }
    findFirst(args) {
        return this.withFluentCall('findFirst', args);
    }
    findFirstOrThrow(args) {
        return this.withFluentCall('findFirstOrThrow', args);
    }
    findMany(args) {
        return this.deferred('findMany', args);
    }
    create(args) {
        return this.deferred('create', args);
    }
    createMany(args) {
        return this.deferred('createMany', args, false);
    }
    createManyAndReturn(args) {
        return this.deferred('createManyAndReturn', args);
    }
    update(args) {
        return this.deferred('update', args);
    }
    updateMany(args) {
        return this.deferred('updateMany', args, false);
    }
    updateManyAndReturn(args) {
        return this.deferred('updateManyAndReturn', args);
    }
    upsert(args) {
        return this.deferred('upsert', args);
    }
    delete(args) {
        return this.deferred('delete', args);
    }
    deleteMany(args) {
        return this.deferred('deleteMany', args, false);
    }
    aggregate(args) {
        return this.deferred('aggregate', args, false);
    }
    groupBy(args) {
        return this.deferred('groupBy', args, false);
    }
    count(args) {
        return this.deferred('count', args, false);
    }
    subscribe(args) {
        return this.doSubscribeStream('subscribe', args);
    }
    stream(args) {
        return this.doSubscribeStream('stream', args);
    }
    doSubscribeStream(method, args) {
        return __awaiter(this, void 0, void 0, function* () {
            // Prisma's `subscribe` and `stream` methods return an async iterable
            // which we need to wrap to process the iteration results
            const iterable = yield this.prisma[this.model][method](args);
            return {
                [Symbol.asyncIterator]: () => {
                    const iter = iterable[Symbol.asyncIterator].bind(iterable)();
                    return {
                        next: () => __awaiter(this, void 0, void 0, function* () {
                            const { done, value } = yield iter.next();
                            const processedValue = value ? yield this.processResultEntity(method, value) : value;
                            return { done, value: processedValue };
                        }),
                        return: () => { var _a; return (_a = iter.return) === null || _a === void 0 ? void 0 : _a.call(iter); },
                        throw: () => { var _a; return (_a = iter.throw) === null || _a === void 0 ? void 0 : _a.call(iter); },
                    };
                },
            };
        });
    }
    /**
     * Processes result entities before they're returned
     */
    processResultEntity(_method, data) {
        return __awaiter(this, void 0, void 0, function* () {
            return data;
        });
    }
    /**
     * Processes query args before they're passed to Prisma.
     */
    preprocessArgs(_method, args) {
        return __awaiter(this, void 0, void 0, function* () {
            return args;
        });
    }
}
exports.DefaultPrismaProxyHandler = DefaultPrismaProxyHandler;
// a marker for filtering error stack trace
const ERROR_MARKER = '__error_marker__';
const customInspect = Symbol.for('nodejs.util.inspect.custom');
/**
 * Makes a Prisma client proxy.
 */
function makeProxy(prisma, modelMeta, makeHandler, name = 'unnamed_enhancer', errorTransformer) {
    const models = Object.keys(modelMeta.models).map((k) => k.toLowerCase());
    const proxy = new Proxy(prisma, {
        get: (target, prop, receiver) => {
            // enhancer metadata
            if (prop === constants_1.PRISMA_PROXY_ENHANCER) {
                return name;
            }
            if (prop === '$transaction') {
                // for interactive transactions, we need to proxy the transaction function so that
                // when it runs the callback, it provides a proxy to the Prisma client wrapped with
                // the same handler
                //
                // TODO: batch transaction is not supported yet, how?
                const $transaction = Reflect.get(target, prop, receiver);
                if ($transaction) {
                    return (input, ...rest) => {
                        if (Array.isArray(input)) {
                            throw new Error('Sequential operations transaction is not supported by ZenStack enhanced Prisma client. Please use interactive transaction instead.');
                        }
                        else if (typeof input !== 'function') {
                            throw new Error('A function value input is expected');
                        }
                        const txFunc = input;
                        return $transaction.bind(target)((tx) => {
                            // create a proxy for the transaction function
                            const txProxy = makeProxy(tx, modelMeta, makeHandler, name + '$tx');
                            // call the transaction function with the proxy
                            return txFunc(txProxy);
                        }, ...rest);
                    };
                }
                else {
                    return $transaction;
                }
            }
            if (prop === '$extends') {
                // Prisma's `$extends` API returns a new client instance, we need to recreate
                // a proxy around it
                const $extends = Reflect.get(target, prop, receiver);
                if ($extends && typeof $extends === 'function') {
                    return (...args) => {
                        const result = $extends.bind(target)(...args);
                        return makeProxy(result, modelMeta, makeHandler, name + '$ext', errorTransformer);
                    };
                }
                else {
                    return $extends;
                }
            }
            if (typeof prop !== 'string' || prop.startsWith('$') || !models.includes(prop.toLowerCase())) {
                // skip non-model fields
                return Reflect.get(target, prop, receiver);
            }
            const propVal = Reflect.get(target, prop, receiver);
            if (!propVal || typeof propVal !== 'object') {
                return propVal;
            }
            return createHandlerProxy(makeHandler(target, prop), propVal, prop, proxy, errorTransformer);
        },
    });
    proxy[customInspect] = `$zenstack_prisma_${prisma._clientVersion}`;
    return proxy;
}
// A proxy for capturing errors and processing stack trace
function createHandlerProxy(handler, origTarget, model, dbOrTx, errorTransformer) {
    return new Proxy(handler, {
        get(target, propKey) {
            if (propKey === '$parent') {
                return dbOrTx;
            }
            const prop = target[propKey];
            if (typeof prop !== 'function') {
                // the proxy handler doesn't have this method, fall back to the original target
                // this can happen for new methods added by Prisma Client Extensions
                return origTarget[propKey];
            }
            // eslint-disable-next-line @typescript-eslint/ban-types
            const origMethod = prop;
            return function (...args) {
                // using proxy with async functions results in messed-up error stack trace,
                // create an error to capture the current stack
                const capture = new Error(ERROR_MARKER);
                // the original proxy returned by the PrismaClient proxy
                const promise = origMethod.apply(handler, args);
                // modify the error stack
                const resultPromise = (0, promise_1.createDeferredPromise)(() => {
                    return new Promise((resolve, reject) => {
                        promise.then((value) => resolve(value), (err) => {
                            if (capture.stack && err instanceof Error) {
                                // save the original stack and replace it with a clean one
                                err.internalStack = err.stack;
                                err.stack = cleanCallStack(capture.stack, model, propKey.toString(), err.message);
                            }
                            if (errorTransformer) {
                                err = errorTransformer ? errorTransformer(err) : err;
                            }
                            reject(err);
                        });
                    });
                });
                // carry over extra fields from the original promise
                for (const [k, v] of Object.entries(promise)) {
                    if (!(k in resultPromise)) {
                        resultPromise[k] = v;
                    }
                }
                return resultPromise;
            };
        },
    });
}
// Filter out @zenstackhq/runtime stack (generated by proxy) from stack trace
function cleanCallStack(stack, model, method, message) {
    // message line
    let resultStack = `Error calling enhanced Prisma method \`${model}.${method}\`: ${message}`;
    const lines = stack.split('\n');
    let foundMarker = false;
    for (let i = 0; i < lines.length; i++) {
        const line = lines[i];
        if (!foundMarker) {
            // find marker, then stack trace lines follow
            if (line.includes(ERROR_MARKER)) {
                foundMarker = true;
            }
            continue;
        }
        // skip leading zenstack and anonymous lines
        if (line.includes('@zenstackhq/runtime') || line.includes('Proxy.<anonymous>')) {
            continue;
        }
        // capture remaining lines
        resultStack += lines
            .slice(i)
            .map((l) => '\n' + l)
            .join();
        break;
    }
    return resultStack;
}
//# sourceMappingURL=proxy.js.map